from pathlib import Path
import pandas as pd


METADATA_SCHEMA = sorted(["Path", "From", "Date"])


def make_email_metadata(email_path: str | Path) -> dict[str, str]:
    """Make an email metadata object by parsing the email contents

    Args:
        email_path:
        email_contents: Contents of an email file (from f.readlines())

    Returns: Dict containing the required metadata
    """
    key_is_set = {key: False for key in METADATA_SCHEMA}
    metadata = {"Path": str(email_path)}
    key_is_set["Path"] = True

    with open(email_path, "r") as f:
        for line in f:
            try:
                k, v = [val.strip() for val in line.split(":", maxsplit=1)]
                if k in METADATA_SCHEMA:
                    metadata[k] = v
                    key_is_set[k] = True
                    if all([v for _, v in key_is_set.items()]):
                        break
            # If it's not a metadata line just pass
            except ValueError:
                pass

    # Validate that all keys have been found
    assert sorted(list(metadata.keys())) == METADATA_SCHEMA
    return metadata


def parse_maildir(path_to_dir: str | Path):
    """Parse all of the emails within the specified maildir box (not recursively)

    Args:
        path_to_dir: Path to maildir

    Returns: MailDir object initialized with email information

    """
    file_list = Path(path_to_dir).glob("*")
    email_metadata = [make_email_metadata(file) for file in file_list]
    return MailDir(email_metadata)


class TopSender:
    """Simple class to store the top sender alongside the first 5 names they used"""

    def __init__(self, email: str, names: list[str]):
        self.email = email
        self.names = names


class MailDir:
    """Representation of the local maildir repository.
    Stores the metadata associated with all local emails.
    """

    KEYS_AND_FUNCS = {
        "Name": lambda df: df["From"].map(lambda x: x.split("<")[0].strip('" ')),
        "Email": lambda df: df["From"].map(lambda x: x.split("<")[1].strip(">")),
        "Date": lambda df: pd.to_datetime(df["Date"]),
    }

    def __init__(self, email_metadata: list[dict[str, str]]):
        if email_metadata:
            self._df = pd.DataFrame(email_metadata)

            for k, func in self.KEYS_AND_FUNCS.items():
                self._df[k] = func(self._df)

        else:
            self._df = pd.DataFrame(
                columns=METADATA_SCHEMA + list(self.KEYS_AND_FUNCS.keys())
            )

    def get_top_n_senders(self, n: int) -> list[TopSender]:
        """Calculate the top n senders and returns their information as a TopSender object
        The TopSender object
        Args:
            n: Number of senders to retrieve

        Returns: list of TopSender objects

        """

        senders = [
            TopSender(
                email, list(self._df.loc[self._df["Email"] == email, "Name"].unique())
            )
            for email in self._df["Email"].value_counts().iloc[0:n].index
        ]

        return senders
